"""
Specification for various utility types used throughout tools and CryoSPARC.
"""

from typing import Dict, List, Literal, Tuple, TypedDict, Union

Shape = Tuple[int, ...]
"""A numpy shape tuple from ndarray.shape"""

DType = Union[str, Tuple[str, Shape]]
"""

    Can just be a single string such as "f4", "3u4" or "O".
    A datatype description of a ndarray entry.

    Can also be the a tuple with a string datatype name and its shape. For
    example, the following dtypes are equivalent.

    - "3u4"
    - "<u4", (3,))
"""

Field = Union[Tuple[str, str], Tuple[str, str, Shape]]
"""
    Description of a column in a numpy array with named fields

    Examples:
    - ("uid", "u8")
    - ("coords", "3f4")
    - ("coords", "<f4", (3,))
"""

Datatype = Literal[
    "exposure",
    "particle",
    "template",
    "volume",
    "volume_multi",
    "mask",
    "live",
    "ml_model",
    "symmetry_candidate",
    "flex_mesh",
    "flex_model",
    "hyperparameter",
    "denoise_model",
    "annotation_model",
]
"""Supported data types for job inputs and outputs."""


LoadableSlots = Union[Literal["default", "passthrough", "all"], List[str]]
"""
Types of slots that may be included when loading an input or an output.
"default" means only slots required as inputs or newly-generated by outputs.
"passthrough" means only unused passthroughs for inputs/outputs. "all" means the
previous two combined. list means a specific list of slots.
"""


# Valid plot file types
TextFormat = Literal["txt", "csv", "html", "json", "xml", "bild", "bld", "log"]
"""
Supported job stream log asset file text formats.
"""

ImageFormat = Literal["pdf", "gif", "jpg", "jpeg", "png", "svg"]
"""
Supported job stream log asset file image formats.
"""

MicrographFormat = Literal["mrc", "mrcs", "stk", "tif", "tiff", "eer", "ecc", "bz2", "cmrcbz2"]
"""Supported micrograph extensions."""

DatasetFormat = Literal["cs", "npy"]
"""Supported extensions for dataset files."""

AssetFormat = Union[TextFormat, ImageFormat]
"""
Supported job stream log asset file formats.
"""

AnyFormat = Union[TextFormat, ImageFormat, MicrographFormat, DatasetFormat]
"""
Supported file extensions.
"""

TextContentType = Literal[
    "text/plain",
    "text/csv",
    "text/html",
    "application/json",
    "application/xml",
    "application/x-troff",
]
"""
Supported job stream log text asset MIME types.
"""

ImageContentType = Literal[
    "application/pdf",
    "image/gif",
    "image/jpeg",
    "image/png",
    "image/svg+xml",
]
"""
Supported job image asset MIME types renderable in the web
"""

MicrographContentType = Literal[
    "image/x-mrc",
    "image/tiff",
    "application/x-eer",
    "application/x-bzip2",
]
"""Supported micrograph content types."""

DatasetContentType = Literal["application/x-cryosparc-dataset", "application/x-numpy"]
"""Supported content types for dataset files."""

AssetContentType = Union[TextContentType, ImageContentType, Literal["application/octet-stream"]]
"""
Supported job asset MIME types.
"""

AnyContentType = Union[
    TextContentType,
    ImageContentType,
    MicrographContentType,
    DatasetContentType,
    Literal["application/octet-stream"],
]
"""
All known and supported content types used by CryoSPARC.
"""

TEXT_CONTENT_TYPES: Dict[TextFormat, TextContentType] = {
    "log": "text/plain",
    "txt": "text/plain",
    "csv": "text/csv",
    "html": "text/html",
    "json": "application/json",
    "xml": "application/xml",
    "bild": "application/x-troff",
    "bld": "application/x-troff",
}

IMAGE_CONTENT_TYPES: Dict[ImageFormat, ImageContentType] = {
    "pdf": "application/pdf",
    "gif": "image/gif",
    "jpeg": "image/jpeg",
    "jpg": "image/jpeg",
    "png": "image/png",
    "svg": "image/svg+xml",
}

MICROGRAPH_CONTENT_TYPES: Dict[MicrographFormat, MicrographContentType] = {
    "mrc": "image/x-mrc",
    "mrcs": "image/x-mrc",
    "stk": "image/x-mrc",
    "tif": "image/tiff",
    "tiff": "image/tiff",
    "eer": "application/x-eer",
    "ecc": "application/x-eer",
    "bz2": "application/x-bzip2",
    "cmrcbz2": "application/x-bzip2",
}

DATASET_CONTENT_TYPES: Dict[DatasetFormat, DatasetContentType] = {
    "cs": "application/x-cryosparc-dataset",  # but usually has numpy format internally
    "npy": "application/x-numpy",
}

ASSET_CONTENT_TYPES: Dict[AssetFormat, AssetContentType] = {**TEXT_CONTENT_TYPES, **IMAGE_CONTENT_TYPES}
ASSET_EXTENSIONS: Dict[AssetContentType, AssetFormat] = {v: k for k, v in ASSET_CONTENT_TYPES.items()}

KNOWN_CONTENT_TYPES: Dict[AnyFormat, AnyContentType] = {
    **TEXT_CONTENT_TYPES,
    **IMAGE_CONTENT_TYPES,
    **MICROGRAPH_CONTENT_TYPES,
    **DATASET_CONTENT_TYPES,
}


# Slot is defined in two classes like this because it's the only way to
# make the ``required`` key optional.
class _Slot(TypedDict):
    """
    :meta private:
    """

    name: str
    dtype: str


class Slot(_Slot, total=False):
    """
    Full slot dictionary specification type for items in the slots=... argument
    when creating inputs or outputs. e.g., ``{"name": "ctf", "dtype": "ctf"}``
    or ``{"name": "background_blob", "dtype": "stat_blob", "required": False}``

    See :py:type:`SlotSpec` for details.

    Attributes:
        name (str): where to find field in a corresponding .cs file e.g.,
            ``"background_blob"``, ``"ctf"``, ``"alignments_class_0"``
        dtype (str): name of known data type. e.g., ``"stat_blob"``, ``"ctf"``,
            ``"alignments3D"``.
        required (bool, optional): Whether this slot is required. Applies to
            input specs only. Defaults to True.

    """

    required: bool


class Datafield(TypedDict):
    """
    Deprecated. Use :py:class:`Slot` instead.
    """

    dtype: str
    prefix: str
    required: bool


SlotSpec = Union[str, Slot, Datafield]
"""
A result slot specification for items in the slots=... argument when creating
inputs or outputs.

In CryoSPARC, all jobs have one or more inputs and outputs. An input or output
has some broad :py:type:`Datatype`, such as ``"exposure"`` or ``"particle"``.
Each input or output also has a list of associated "low-level" results created
at various stages of processing, such as ``"location"`` for picked particles and
``blob`` for extracted particles. A slot represents one of these low-level
results.

In the CryoSPARC interface, open a job's "Inputs" or "Outputs" tab to see the
kinds of slots available. You may also download an output and load it with
:py:type:`~cryosparc.dataset.Dataset` to inspect the infomation encoded in its
results.

Provide each slot as either a string representing a name and result type, or a
full dictionary specification.

A string in the format ``"<slot>"`` is a shortcut for ``{"name": "<slot>",
"dtype": "<known slot type>", "required": True}``.

A string in the format ``"?<slot>"`` is a shortcut for ``{"name": "<slot>",
"dtype": "<known slot type>", "required": False}`` (input slots only).

Example strings::

    "ctf"
    "micrograph_blob"
    "?background_blob"

Example equivalent full specifications::

    {"name": "ctf", "dtype": "ctf"}
    {"name": "micrograph_blob", "dtype": "micrograph_blob", "required": True}
    {"name": "background_blob", "dtype": "stat_blob", "required": False}

Use the full specification when the ``dtype`` cannot be inferred from the
``name`` string because it is dynamic. For example, 3D Variability job
``particles`` outputs have slots named ``"components_mode_X"`` with dtype
``"components"`` where ``X`` is a mode number::

    [
        "blob",
        "?locations",
        {"name": "components_mode_0", "dtype": "components"},
        {"name": "components_mode_1", "dtype": "components", "required": False},
        {"name": "components_mode_2", "dtype": "components", "required": False},
    ]

Note that the ``required`` key only applies to input slots.
"""


class JobSection(TypedDict):
    """
    Deprecated. Use :py:class:`~cryosparc.models.job_register.JobRegister`
    instead.
    """

    name: str
    title: str
    description: str
    contains: List[str]



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Literal, Optional

from pydantic import BaseModel, Field

from .job import JobStatus


class Notification(BaseModel):
    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: Optional[str] = None
    """
    """
    job_uid: Optional[str] = None
    """
    """
    message: str
    """
    """
    progress_pct: Optional[float] = None
    """
    Progress percentage (0-100).
    """
    active: bool = True
    """
    """
    status: Optional[Literal["success", "primary", "warning", "danger"]] = "success"
    """
    """
    icon: str = "flag"
    """
    """
    hide: bool = False
    """
    """
    job_status: Optional[JobStatus] = None
    """
    """
    ttl_seconds: Optional[int] = 7
    """
    Time-to-live in seconds for automatic deactivation.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .asset import GridFSAsset


class CheckpointEvent(BaseModel):
    """
    An event type indicating a checkpoint in the job's execution.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    Project UID that this event belongs to.
    """
    job_uid: str
    """
    Job UID that this event belongs to.
    """
    cpumem_mb: Optional[float] = None
    """
    CPU memory used by the job process at the time of the event, in MB.
    """
    avail_mb: Optional[float] = None
    """
    Available worker system memory at the time of the event, in MB.
    """
    flags: List[str] = []
    """
    List of flags associated with this event.
    """
    meta: Dict[str, Any] = {}
    """
    Metadata associated with this event.
    """
    type: Literal["checkpoint"]
    """
    Checkpoint events always have type 'checkpoint'.
    """


class Event(BaseModel):
    """
    Base class for all event types.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    Project UID that this event belongs to.
    """
    job_uid: str
    """
    Job UID that this event belongs to.
    """
    cpumem_mb: Optional[float] = None
    """
    CPU memory used by the job process at the time of the event, in MB.
    """
    avail_mb: Optional[float] = None
    """
    Available worker system memory at the time of the event, in MB.
    """
    flags: List[str] = []
    """
    List of flags associated with this event.
    """
    meta: Dict[str, Any] = {}
    """
    Metadata associated with this event.
    """


class ImageEvent(BaseModel):
    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    Project UID that this event belongs to.
    """
    job_uid: str
    """
    Job UID that this event belongs to.
    """
    cpumem_mb: Optional[float] = None
    """
    CPU memory used by the job process at the time of the event, in MB.
    """
    avail_mb: Optional[float] = None
    """
    Available worker system memory at the time of the event, in MB.
    """
    flags: List[str] = []
    """
    List of flags associated with this event.
    """
    meta: Dict[str, Any] = {}
    """
    Metadata associated with this event.
    """
    type: Literal["image"]
    """
    Image events always have type 'image'.
    """
    text: str
    """
    Text description of the image event.
    """
    imgfiles: List[GridFSAsset] = []
    """
    List of image assets associated with the event.
    """


class InteractiveGridFSAsset(BaseModel):
    """
    Basic information about an uploaded data asset stored in GridFS.
    """

    fileid: str = "000000000000000000000000"
    """
    """
    filename: str
    """
    File name, e.g,. "image.png"
    """
    filetype: str
    """
    File format extension, e.g., "png"
    """


class InteractiveImgfile(BaseModel):
    """
    An image file associated with an interactive event, along with its components.
    """

    imgfiles: List[GridFSAsset]
    """
    List of image assets.
    """
    components: List[int] = []
    """
    List of data component indices which this image file represents.
    """


class InteractiveEvent(BaseModel):
    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    Project UID that this event belongs to.
    """
    job_uid: str
    """
    Job UID that this event belongs to.
    """
    cpumem_mb: Optional[float] = None
    """
    CPU memory used by the job process at the time of the event, in MB.
    """
    avail_mb: Optional[float] = None
    """
    Available worker system memory at the time of the event, in MB.
    """
    flags: List[str] = []
    """
    List of flags associated with this event.
    """
    meta: Dict[str, Any] = {}
    """
    Metadata associated with this event.
    """
    type: Literal["interactive"]
    """
    Interactive events always have type 'interactive'.
    """
    subtype: str = "3dscatter"
    """
    Subtype of interactive event, e.g. '3dscatter'.
    """
    text: str
    """
    Text description of the interactive event.
    """
    datafile: InteractiveGridFSAsset
    """
    Data asset associated with the interactive event.
    """
    preview_imgfiles: List[InteractiveImgfile] = []
    """
    List of preview image files associated with the interactive event.
    """
    components: List[int] = []
    """
    List of data component indices included in this interactive event.
    """


class TextEvent(BaseModel):
    """
    An event with only text and no additional image or interactive data. May
    have "text", "warning" or "error" type.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    Project UID that this event belongs to.
    """
    job_uid: str
    """
    Job UID that this event belongs to.
    """
    cpumem_mb: Optional[float] = None
    """
    CPU memory used by the job process at the time of the event, in MB.
    """
    avail_mb: Optional[float] = None
    """
    Available worker system memory at the time of the event, in MB.
    """
    flags: List[str] = []
    """
    List of flags associated with this event.
    """
    meta: Dict[str, Any] = {}
    """
    Metadata associated with this event.
    """
    type: Literal["text", "warning", "error"]
    """
    Type of text event: 'text', 'warning', or 'error'.
    """
    text: str
    """
    Text content of the event.
    """



from . cimport dataset
from . cimport lz4
from libc.stdint cimport uint8_t, uint16_t, uint64_t
from cpython.ref cimport PyObject, Py_XINCREF, Py_XDECREF
from cpython.mem cimport PyMem_Realloc, PyMem_Free

cdef int LZ4_ACCELERATION = 1000


# Mirror of equivalent C-datatype enumeration
cpdef enum DsetType:
    T_F32 = 1
    T_F64 = 2
    T_C32 = 3
    T_C64 = 4
    T_I8 = 5
    T_I16 = 6
    T_I32 = 7
    T_I64 = 8
    T_U8 = 9
    T_U16 = 10
    T_U32 = 11
    T_U64 = 12
    T_STR = 13
    T_OBJ = 14


cdef class Data:
    cdef dataset.Dset _handle

    def __cinit__(self, other = None):
        cdef Data othr
        self._handle = 0
        if isinstance(other, Data):
            # copy constructor
            othr = <Data> other
            self._handle = dataset.dset_copy(othr._handle)
            othr._increfs()
        elif other:
            # Initialize with a numeric handle
            self._handle = <dataset.Dset> other
        else:
            self._handle = dataset.dset_new()

        if self._handle == 0:
            raise MemoryError("Could not allocate dataset")

    def __dealloc__(self):
        if self._handle:
            self._decrefs()
            dataset.dset_del(self._handle)

    def _increfs(self):
        # Increment reference counts for all Python object fields.
        # Call this after making a copy of this dataset.
        nrow = dataset.dset_nrow(self._handle)
        for i in xrange(dataset.dset_ncol(self._handle)):
            with nogil:
                colkey = dataset.dset_key(self._handle, i)
                coltype = dataset.dset_type(self._handle, colkey)
                if coltype != DsetType.T_OBJ:
                    continue
                mem = <PyObject **> (dataset.dset_get(self._handle, colkey))

            for j in xrange(nrow):
                Py_XINCREF(mem[j])

    def _decrefs(self):
        # Decrement reference counts for all Python object fields.
        # Call this after destroying this dataset.
        nrow = dataset.dset_nrow(self._handle)
        for i in xrange(dataset.dset_ncol(self._handle)):
            with nogil:
                colkey = dataset.dset_key(self._handle, i)
                coltype = dataset.dset_type(self._handle, colkey)
                if coltype != DsetType.T_OBJ:
                    continue
                mem = <PyObject **> (dataset.dset_get(self._handle, colkey))

            for j in xrange(nrow):
                Py_XDECREF(mem[j])

    def innerjoin(self, str key, Data other):
        cdef Data data = Data(dataset.dset_innerjoin(key.encode(), self._handle, other._handle))
        data._increfs()
        return data

    def totalsz(self):
        return dataset.dset_totalsz(self._handle)

    def ncol(self):
        return dataset.dset_ncol(self._handle)

    def nrow(self):
        return dataset.dset_nrow(self._handle)

    def key(self, int index):
        return dataset.dset_key(self._handle, index).decode()

    def type(self, str field):
        return dataset.dset_type(self._handle, field.encode())

    def has(self, str field):
        return self.type(field) > 0

    def addrows(self, int num):
        if not dataset.dset_addrows(self._handle, num):
            raise MemoryError("Could not add rows to dataset")

    def addcol_scalar(self, str field, int dtype):
        if not dataset.dset_addcol_scalar(self._handle, field.encode(), dtype):
            raise MemoryError("Could not add column to dataset")

    def addcol_array(self, str field, int dtype, tuple shape):
        if not 0 < len(shape) <= 3:
            raise ValueError("Shape size must be between 0 and 3")

        cdef uint16_t c_shape[3]
        c_shape[0] = 0; c_shape[1] = 0; c_shape[2] = 0
        for i in xrange(len(shape)):
            c_shape[i] = <uint16_t> shape[i]

        if not dataset.dset_addcol_array(self._handle, field.encode(), dtype, c_shape):
            raise MemoryError("Could not add column to dataset")

    def getshp(self, str colkey):
        cdef list shp = []
        cdef uint64_t val = dataset.dset_getshp(self._handle, colkey.encode())
        cdef uint16_t dim0 = <uint16_t> (val & 0xFFFF)
        cdef uint16_t dim1 = <uint16_t> ((val >> 16) & 0xFFFF)
        cdef uint16_t dim2 = <uint16_t> ((val >> 32) & 0xFFFF)
        if dim0: shp.append(<int> dim0)
        if dim1: shp.append(<int> dim1)
        if dim2: shp.append(<int> dim2)
        return tuple(shp)

    def getbuf(self, str colkey):
        cdef void *mem
        cdef size_t size
        cdef bytes colkey_b = colkey.encode()
        cdef const char *colkey_c = colkey_b
        with nogil:
            mem = dataset.dset_get(self._handle, colkey_c)
            size = dataset.dset_getsz(self._handle, colkey_c)
        return None if size == 0 else <unsigned char [:size]> mem

    def getstr(self, str col, size_t index):
        return dataset.dset_getstr(self._handle, col.encode(), index)  # returns bytes

    def tocstrs(self, str col):
        # Convert Python strings to C strings in the given object column
        cdef bytes colkey = col.encode()
        cdef int prevtype = dataset.dset_type(self._handle, colkey)
        cdef size_t nrow = dataset.dset_nrow(self._handle)
        cdef str pystr
        cdef bytes pybytes

        if prevtype != T_OBJ or not dataset.dset_changecol(self._handle, colkey, T_STR):
            return False

        cdef PyObject **pycol = <PyObject **> dataset.dset_get(self._handle, colkey)
        for i in xrange(nrow):
            pycol = <PyObject **> dataset.dset_get(self._handle, colkey)
            pystr = <str> pycol[i]
            pybytes = pystr.encode()
            Py_XDECREF(pycol[i])  # so string is deallocated
            pycol[i] = NULL  # so that strfree not attempted
            if not dataset.dset_setstr(self._handle, colkey, i, pybytes, len(pybytes)):
                raise MemoryError("Could not convert strings")

        return True

    def topystrs(self, str col):
        # Convert C strings to Python strings in the given column
        cdef bytes colkey = col.encode()
        cdef int prevtype = dataset.dset_type(self._handle, colkey)
        cdef size_t nrow = dataset.dset_nrow(self._handle)
        cdef size_t *pycol = <size_t *> dataset.dset_get(self._handle, colkey)
        cdef PyObject **pystrcol = <PyObject **> pycol
        cdef dict strcache = dict()
        cdef char *cstr
        cdef bytes cbytes
        cdef str pystr

        if prevtype != T_STR:
            return False

        for i in xrange(nrow):
            if pycol[i] in strcache:
                pystr = strcache[pycol[i]]
            else:
                cbytes = dataset.dset_getstr(self._handle, colkey, i)
                pystr = cbytes.decode()
                strcache[pycol[i]] = pystr

            pystrcol[i] = <PyObject *> pystr
            Py_XINCREF(<PyObject *> pystr)

        return bool(dataset.dset_changecol(self._handle, colkey, T_OBJ))

    def stralloc(self, str val):
        cdef uint64_t idx
        cdef bytes pybytes = val.encode()
        if not dataset.dset_stralloc(self._handle, pybytes, len(pybytes), &idx):
            raise MemoryError("Could not allocate string in dataset")
        return <int> idx

    def dump(self):
        cdef void *mem
        cdef size_t size
        with nogil:
            mem = dataset.dset_dump(self._handle)
            size = dataset.dset_totalsz(self._handle)
        return <unsigned char [:size]> mem

    def dumpstrheap(self):
        cdef void *mem
        cdef size_t size
        with nogil:
            mem = dataset.dset_strheap(self._handle)
            size = dataset.dset_strheapsz(self._handle)
        return <unsigned char [:size]> mem

    def setstrheap(self, bytes heap):
        if not dataset.dset_setstrheap(self._handle, <const char *> heap, len(heap)):
            raise MemoryError("Could not set string heap in dataset")

    def defrag(self, bint realloc_smaller):
        if not dataset.dset_defrag(self._handle, realloc_smaller):
            raise MemoryError("Could not defrag dataset")

    def dumptxt(self, bint dump_data = 0):
        dataset.dset_dumptxt(self._handle, dump_data)

    def handle(self):
        return self._handle


cdef class Stream:
    # Helper class for initializing a dataset from a compressed stream or
    # generating a compressed stream for that dataset.
    #
    # WARNING: Methods here recycle result buffers to minimize allocations. If
    # results are not consumed prior to future calls to this class, the contents
    # will be overwritten.
    cdef Data data
    cdef char *buf  # internal compression buffer
    cdef uint64_t *aux  # internal string index array
    cdef size_t bufsz
    cdef size_t auxsz

    def __cinit__(self, Data data = None):
        self.data = data
        self.buf = NULL
        self.aux = NULL
        self.bufsz = 0
        self.auxsz = 0

    def __dealloc__(self):
        if self.buf != NULL:
            PyMem_Free(self.buf)
            self.buf = NULL
        if self.aux != NULL:
            PyMem_Free(self.aux)
            self.aux = NULL

    def _ensure_buf(self, int min_len):
        cdef size_t sz = min_len
        if self.bufsz < sz:
            self.buf = <char *> PyMem_Realloc(<void *> self.buf, sz)
            self.bufsz = sz
        return <int> self.bufsz

    def _ensure_aux(self, int min_sz):
        cdef size_t sz = min_sz
        if self.auxsz < sz:
            self.aux = <uint64_t *> PyMem_Realloc(<void *> self.aux, sz)
            self.auxsz = sz
        return <int> self.auxsz

    def cast_objs_to_strs(self):
        # change all T_OBJ column types to T_STR (underlying data unaffected)
        cdef dataset.Dset handle = self.data._handle
        cdef int ncol = self.data.ncol()
        cdef int coltype
        cdef const char *colkey

        with nogil:
            for i in xrange(ncol):
                colkey = dataset.dset_key(handle, i)
                coltype = dataset.dset_type(handle, colkey)
                if coltype == T_OBJ:
                    dataset.dset_changecol(handle, colkey, T_STR)

    def stralloc_col(self, str col):
        # Allocate C strings for every Python string in the given object column
        # Returns 0 if the column does not have object type.
        cdef dataset.Dset handle = self.data._handle
        cdef bytes colkey = col.encode()
        cdef int coltype = dataset.dset_type(handle, colkey)
        if coltype != T_OBJ:
            return 0  # invalid column

        # Convert to C strings and compress index array instead
        cdef PyObject **coldata = <PyObject **> dataset.dset_get(handle, colkey)
        cdef uint64_t nrow = dataset.dset_nrow(handle)
        cdef size_t sz = dataset.dset_getsz(handle, colkey)

        self._ensure_aux(sz)

        cdef uint64_t idx
        cdef int allocres
        cdef str pystr
        cdef bytes pybytes

        for i in xrange(nrow):
            pystr = <str> coldata[i]
            pybytes = pystr.encode()
            allocres = dataset.dset_stralloc(handle, pybytes, len(pybytes), &idx)
            if allocres == 0:
                raise MemoryError("Could not allocate string in dataset")
            elif allocres == 2:
                # dataset reallocated, coldata must be retrieved
                coldata = <PyObject **> dataset.dset_get(handle, colkey)
            self.aux[i] = idx

        return <unsigned char [:sz]> (<unsigned char *> self.aux)

    def compress_col(self, str col):
        # Use when multiple compression calls are required to minimize total
        # number of allocations.
        #
        # Overwrites (and potentially re-allocates) internal buffer each time
        # it's called so only the latest return value is valid memory (until the
        # instance is garbage collected).
        #
        # If the column is has type T_OBJ, allocates each python string as
        # a C string in the string heap and compresses the resulting array of
        # indexes into the C string.
        #
        # Returns array memoryview with compressed data
        cdef dataset.Dset handle = self.data._handle
        cdef colkey = col.encode()
        cdef int coltype = dataset.dset_type(handle, colkey)

        if coltype == 0:
            return 0  # invalid column

        cdef uint64_t nrow = dataset.dset_nrow(handle)
        cdef size_t sz = dataset.dset_getsz(handle, colkey)
        cdef uint64_t idx
        cdef int allocres
        cdef str pystr
        cdef bytes pybytes
        cdef PyObject **coldata
        cdef unsigned char [:] data

        if coltype == T_OBJ:
            # Convert to C strings and compress index array instead
            self._ensure_aux(sz)
            coldata = <PyObject **> dataset.dset_get(handle, colkey)

            for i in xrange(nrow):
                pystr = <str> coldata[i]
                pybytes = pystr.encode()
                allocres = dataset.dset_stralloc(handle, pybytes, len(pybytes), &idx)
                if allocres == 0:
                    raise MemoryError("Could not allocate string in dataset for compression")
                elif allocres == 2:
                    # dataset reallocated, coldata must be retrieved
                    coldata = <PyObject **> dataset.dset_get(handle, colkey)
                self.aux[i] = idx

            data = <unsigned char [:sz]> (<unsigned char *> self.aux)
        else:
            data = <unsigned char [:sz]> (<unsigned char *> dataset.dset_get(handle, colkey))

        return self.compress(data)

    def compress_numpy(self, arr):
        cdef size_t sz = arr.size * arr.itemsize
        cdef size_t arr_ptr_val = arr.ctypes.data
        cdef void *arr_ptr = <void *> arr_ptr_val
        return self.compress(<unsigned char [:sz]> arr_ptr)

    def compress(self, unsigned char [:] data):
        cdef int sz = data.size
        cdef int max_compressed_sz = lz4.LZ4_compressBound(sz)

        self._ensure_buf(max_compressed_sz)

         # Call compress. Write final length into compressed_sz
        cdef int compressed_sz = lz4.LZ4_compress_fast(
            <const char *> &data[0],
            self.buf,
            sz,
            max_compressed_sz,
            LZ4_ACCELERATION
        )
        if compressed_sz > 0:
            return <char [:compressed_sz]> self.buf

        raise ValueError(f"Could not compress (error {compressed_sz})")

    def decompress_col(self, str col, bytes data):
        cdef void *mem
        cdef size_t size
        cdef bytes colkey_b = col.encode()
        cdef const char *colkey_c = colkey_b
        with nogil:
            mem = dataset.dset_get(self.data._handle, colkey_c)
            size = dataset.dset_getsz(self.data._handle, colkey_c)
        if mem == NULL:
            raise ValueError(f"Invalid column {col}")
        else:
            return self.decompress(data, <size_t> mem, size)

    def decompress_numpy(self, bytes data, arr):
        cdef size_t dstptr = arr.ctypes.data
        cdef int size = arr.size
        cdef int itemsize = arr.itemsize
        return self.decompress(data, dstptr, size * itemsize)

    def decompress(self, bytes data, size_t dstptr = 0, int dstsz = 0):
        cdef const char *compressed = data
        cdef int compressed_sz = len(data)
        cdef char *uncompressed = <char *> dstptr
        if dstsz <= 0:
            raise ValueError("Decompression buffer size must be > 0")

        if dstptr == 0:
            self._ensure_buf(dstsz)
            uncompressed = self.buf

        cdef int uncompressed_sz = lz4.LZ4_decompress_safe(compressed, uncompressed, compressed_sz, dstsz)
        if uncompressed_sz >= 0:
            return <char [:uncompressed_sz]> uncompressed

        raise ValueError(f"Could not decompress (error {uncompressed_sz})")



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from .job import JobStatus, RunError
from .session_params import LiveAbinitParams, LiveClass2DParams, LivePreprocessingParams, LiveRefineParams
from .session_spec import SessionStatus
from .signature import ImportSignature
from .workspace import JobGroup, WorkspaceStats


class AthenaVolumeUploadParams(BaseModel):
    """
    :meta private:
    """

    type: Literal["refinement"] = "refinement"
    """
    """
    name: str
    """
    """
    path_rel: str
    """
    """
    symmetry: str = "C1"
    """
    """
    psize_A: Union[float, Literal["inf", "-inf", "nan"]] = 1.0
    """
    """
    res_A: Union[float, Literal["inf", "-inf", "nan"], None] = None
    """
    """


class DataManagementStat(BaseModel):
    """
    Statistics about a specific set of session data and associated storage statuses
    """

    status: Literal["active", "archiving", "archived", "deleted", "deleting", "missing", "calculating"] = "active"
    """
    """
    prev_status: Optional[
        Literal["active", "archiving", "archived", "deleted", "deleting", "missing", "calculating"]
    ] = None
    """
    """
    size: int = 0
    """
    """


class DataManagementStats(BaseModel):
    """
    Statistics and storage statuses for data available in a session
    """

    raw: DataManagementStat = DataManagementStat()
    """
    """
    micrographs: DataManagementStat = DataManagementStat()
    """
    """
    thumbnails: DataManagementStat = DataManagementStat()
    """
    """
    particles: DataManagementStat = DataManagementStat()
    """
    """
    metadata: DataManagementStat = DataManagementStat()
    """
    """


class ECLSessionProperties(BaseModel):
    """
    :meta private:
    """

    do_athena_volume_upload: bool = False
    """
    """
    athena_volume_upload_params: Optional[AthenaVolumeUploadParams] = None
    """
    """


class ExposureGroup(BaseModel):
    """
    Full exposure group defintion, not all properties are externally editable
    """

    ignore_exposures: bool = False
    """
    Set to disable new incoming exposures in this group from being processed
    """
    gainref_path: Optional[str] = None
    """
    Path to gain reference file for exposures in this group
    """
    defect_path: Optional[str] = None
    """
    Path to defect file for exposures in this group
    """
    file_engine_recursive: bool = False
    """
    Whether to recursively watch subdirectories when looking for new exposures
    """
    file_engine_watch_path_abs: str = "/"
    """
    Absolute path to directory to watch for new exposure files
    """
    file_engine_enable: bool = False
    """
    Whether to enable looking for new exposures in this group
    """
    file_engine_filter: str = "*"
    """
    Glob pattern to filter files when looking for new exposures
    """
    file_engine_interval: int = 10
    """
    Interval in seconds between checks for new exposure files
    """
    file_engine_min_file_size: int = 0
    """
    Minimum file size in bytes for new exposure files
    """
    file_engine_min_modified_time_delta: int = 0
    """
    Ignore new exposure files that have been modified within this many seconds.
    Set to prevent processing from starting on incomplete files.
    """
    exp_group_id: int = 1
    """
    Assigned unique ID for this exposure group within the session
    """
    num_exposures_found: int = 0
    """
    Number of exposures found in this group so far
    """
    file_engine_strategy: Literal["entity", "timestamp", "eclathena"] = "entity"
    """
    Strategy for detecting new exposures within this group
    """
    final: bool = False
    """
    Whether this exposure group is finalized and cannot be edited further
    """
    is_any_eer: bool = False
    """
    Whether this exposure group contains any EER exposures
    """


class ExposureGroupUpdate(BaseModel):
    """
    Public editable properties for exposure group
    """

    ignore_exposures: bool = False
    """
    Set to disable new incoming exposures in this group from being processed
    """
    gainref_path: Optional[str] = None
    """
    Path to gain reference file for exposures in this group
    """
    defect_path: Optional[str] = None
    """
    Path to defect file for exposures in this group
    """
    file_engine_recursive: bool = False
    """
    Whether to recursively watch subdirectories when looking for new exposures
    """
    file_engine_watch_path_abs: str = "/"
    """
    Absolute path to directory to watch for new exposure files
    """
    file_engine_enable: bool = False
    """
    Whether to enable looking for new exposures in this group
    """
    file_engine_filter: str = "*"
    """
    Glob pattern to filter files when looking for new exposures
    """
    file_engine_interval: int = 10
    """
    Interval in seconds between checks for new exposure files
    """
    file_engine_min_file_size: int = 0
    """
    Minimum file size in bytes for new exposure files
    """
    file_engine_min_modified_time_delta: int = 0
    """
    Ignore new exposure files that have been modified within this many seconds.
    Set to prevent processing from starting on incomplete files.
    """


class LiveComputeResources(BaseModel):
    """
    Details about a Live session's scheduler compute resource allocation
    """

    phase_one_lane: Optional[str] = None
    """
    """
    phase_one_gpus: int = 1
    """
    """
    phase_two_lane: Optional[str] = None
    """
    """
    phase_two_gpus: int = 1
    """
    """
    phase_two_ssd: bool = True
    """
    """
    auxiliary_lane: Optional[str] = None
    """
    """
    auxiliary_gpus: int = 1
    """
    """
    auxiliary_ssd: bool = True
    """
    """
    priority: int = 0
    """
    Priority of the session's jobs in the scheduler queue (0-100).
    Higher values indicate higher priority.
    """
    phase_one_workers_per_gpu: int = 1
    """
    Number of preprocessing workers that can share the same GPU
    """


class LiveVolumeInfo(BaseModel):
    """
    Information about a volume generated during Live Ab-Initio Reconstruction
    or Streaming Refinement
    """

    vol_gname: str
    """
    """
    fileid: Optional[str] = None
    """
    """
    selected: bool = False
    """
    """
    num_particles: Optional[int] = None
    """
    """


class Phase2ParticleOutputInfo(BaseModel):
    """
    Information about particle output from 2D classification
    """

    path: Optional[str] = None
    """
    """
    count: int = 0
    """
    """
    fields: List[str] = []
    """
    """


class Threshold(BaseModel):
    """
    Statistics-based threshold details
    """

    min: Optional[float] = None
    """
    """
    max: Optional[float] = None
    """
    """
    value: Optional[float] = None
    """
    """


class RangeThreshold(BaseModel):
    """
    Picking range threshold parameters
    """

    min: Optional[float] = None
    """
    """
    max: Optional[float] = None
    """
    """
    value: Optional[float] = None
    """
    """
    min_value: Optional[float] = None
    """
    """
    max_value: Optional[float] = None
    """
    """


class PickingThresholds(BaseModel):
    """
    Picking thresholds for various particle picking statistics
    """

    manual_ncc_score: Threshold = Threshold()
    """
    """
    manual_power: RangeThreshold = RangeThreshold()
    """
    """
    blob_ncc_score: Threshold = Threshold()
    """
    """
    blob_power: RangeThreshold = RangeThreshold()
    """
    """
    template_ncc_score: Threshold = Threshold()
    """
    """
    template_power: RangeThreshold = RangeThreshold()
    """
    """
    deep_ncc_score: Threshold = Threshold()
    """
    :meta private:
    """
    deep_power: RangeThreshold = RangeThreshold()
    """
    :meta private:
    """


class RtpChild(BaseModel):
    """
    Information about a Live worker job
    """

    uid: str
    """
    """
    status: JobStatus
    """
    """
    rtp_handle_func: Literal[
        "handle_template_creation_class2D", "phase2_class2D_handle", "phase2_abinit_handle", "phase2_refine_handle"
    ]
    """
    """


class RtpWorkerState(BaseModel):
    status: JobStatus
    """
    """
    errors: List[RunError] = []
    """
    """


class SessionLastAccessed(BaseModel):
    """
    Details about the last time a session was accessed by a user
    """

    name: str = ""
    """
    username of last user to access the session
    """
    accessed_at: datetime.datetime = datetime.datetime(1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    timestamp of last access
    """


class SessionAttribute(BaseModel):
    """
    Information about a session attribute, used for filtering and display
    """

    name: str
    """
    """
    title: str
    """
    """
    min: Optional[float] = None
    """
    """
    max: Optional[float] = None
    """
    """
    round: int = 0
    """
    """


class TemplateClassInfo(BaseModel):
    """
    Information about a template class used in 2D classification
    """

    class_idx: int
    """
    """
    fileid: str
    """
    """
    res_A: Union[float, Literal["inf", "-inf", "nan"]]
    """
    """
    selected: bool = False
    """
    """
    num_particles_selected: int = 0
    """
    """
    num_particles_total: int = 0
    """
    """
    mean_prob: Union[float, Literal["inf", "-inf", "nan"]] = 0.0
    """
    """
    class_ess: Union[float, Literal["inf", "-inf", "nan"]] = 0.0
    """
    """


class SessionStats(BaseModel):
    """
    Statistics about the session's processing progress
    """

    total_exposures: int = 0
    """
    """
    total_queued: int = 0
    """
    """
    total_seen: int = 0
    """
    """
    total_in_progress: int = 0
    """
    """
    total_thumbs: int = 0
    """
    """
    total_test: int = 0
    """
    """
    total_accepted: int = 0
    """
    """
    total_rejected: int = 0
    """
    """
    total_failed: int = 0
    """
    """
    total_ready: int = 0
    """
    """
    average_manual_picks: int = 0
    """
    """
    total_blob_picks: int = 0
    """
    """
    total_template_picks: int = 0
    """
    """
    total_deep_picks: int = 0
    """
    """
    total_manual_picks: int = 0
    """
    """
    total_extracted_particles_blob: int = 0
    """
    """
    total_extracted_particles_template: int = 0
    """
    """
    total_extracted_particles_manual: int = 0
    """
    """
    total_extracted_particles_deep: int = 0
    """
    """
    total_extracted_particles: int = 0
    """
    """
    total_manual_picked_exposures: int = 0
    """
    """
    gsfsc: Union[float, Literal["inf", "-inf", "nan"]] = 0.0
    """
    """
    frames: int = 0
    """
    """
    nx: int = 0
    """
    """
    ny: int = 0
    """
    """
    manual_rejected: int = 0
    """
    """
    avg_movies_found_per_hour: int = 0
    """
    """
    avg_movies_ready_per_hour: int = 0
    """
    """
    avg_movies_accepted_per_hour: int = 0
    """
    """
    avg_particles_extracted_per_mic: int = 0
    """
    """
    avg_particles_extracted_per_hour: int = 0
    """
    """


class SessionBuildError(BaseModel):
    type: str
    """
    type of error, e.g. 'missing', 'invalid', etc.
    """
    loc: List[Union[str, int]]
    """
    path to the invalid property
    """
    input: Any
    """
    value of the invalid property, must be serializable
    """
    input_type: str
    """
    type name of the input value
    """


class Session(BaseModel):
    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    dumped_at: Optional[datetime.datetime] = None
    """
    When the model was last dumped to disk
    """
    last_dumped_version: Optional[str] = None
    """
    The version of CryoSPARC last dumped at
    """
    autodump: bool = True
    """
    Whether the model was updated recently and must be dumped

    :meta private:
    """
    uid: str
    """
    Workspace unique ID, e.g., 'W1'.
    """
    project_uid: str
    """
    Project unique ID for , e.g., 'P1'.
    """
    workspace_type: Literal["live"] = "live"
    """
    Always 'live' for Live sessions
    """
    created_by_user_id: Optional[str] = None
    """
    User ID that created the workspace.
    """
    title: Optional[str] = None
    """
    """
    description: Optional[str] = None
    """
    """
    created_by_job_uid: Optional[str] = None
    """
    """
    tags: List[str] = []
    """
    """
    starred_by: List[str] = []
    """
    List of user IDs who have starred the workspace.
    """
    deleted: bool = False
    """
    """
    deleting: bool = False
    """
    """
    last_accessed: Optional[SessionLastAccessed] = None
    """
    """
    workspace_stats: WorkspaceStats = WorkspaceStats()
    """
    """
    imported_at: Optional[datetime.datetime] = None
    """
    """
    groups: List[JobGroup] = []
    """
    """
    session_uid: str
    """
    Unique identifier for the session within the project, e.g. 'S1'
    """
    session_dir: str
    """
    Path to session directory on disk, relative to project directory.
    Usually the same as ``session_uid``.
    """
    status: SessionStatus = "paused"
    """
    Current activity status of the session
    """
    failed_at: List[datetime.datetime] = []
    """
    when session entered failed state
    """
    running_at: List[datetime.datetime] = []
    """
    when start_session() is called (when built jobs are placed into the scheduler queue)
    """
    paused_at: List[datetime.datetime] = []
    """
    times when session was paused
    """
    completed_at: Optional[datetime.datetime] = None
    """
    when all jobs inside session are either completed, failed, or killed (i.e. no running/building/waiting). At least one job must've completed successfully for this field to be populated.
    """
    cleared_at: Optional[datetime.datetime] = None
    """
    when experiment was last cleared
    """
    elapsed_time: float = 0.0
    """
    Amount of time (in seconds) the session spent in running state.
    Only updated when session is paused.
    """
    parameter_version: int = 0
    """
    Version of session parameters in use.
    """
    params: LivePreprocessingParams = LivePreprocessingParams()
    """
    Session parameters for preprocessing. Reconstruction parameters are stored separately.
    """
    attributes: List[SessionAttribute] = [
        SessionAttribute(name="found_at", title="Timestamp", min=None, max=None, round=0),
        SessionAttribute(name="check_at", title="Check Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(name="motion_at", title="Motion Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(name="thumbs_at", title="Thumbs Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(name="ctf_at", title="CTF Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(name="pick_at", title="Pick Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(name="extract_at", title="Extract Stage Completed At", min=None, max=None, round=0),
        SessionAttribute(
            name="manual_extract_at", title="Manual Extract Stage Completed At", min=None, max=None, round=0
        ),
        SessionAttribute(name="ready_at", title="Exposure Ready At", min=None, max=None, round=0),
        SessionAttribute(name="total_motion_dist", title="Total Motion (pix)", min=None, max=None, round=2),
        SessionAttribute(name="max_intra_frame_motion", title="Max In-Frame Motion", min=None, max=None, round=3),
        SessionAttribute(name="average_defocus", title="Defocus Avg. (Å)", min=None, max=None, round=0),
        SessionAttribute(name="defocus_range", title="Defocus Range (Å)", min=None, max=None, round=0),
        SessionAttribute(name="astigmatism_angle", title="Astigmatism Angle (deg)", min=None, max=None, round=1),
        SessionAttribute(name="astigmatism", title="Astigmatism", min=None, max=None, round=2),
        SessionAttribute(name="phase_shift", title="Phase Shift (deg)", min=None, max=None, round=1),
        SessionAttribute(name="ctf_fit_to_A", title="CTF Fit (Å)", min=None, max=None, round=3),
        SessionAttribute(name="ice_thickness_rel", title="Relative Ice Thickness", min=None, max=None, round=3),
        SessionAttribute(name="df_tilt_angle", title="Sample Tilt (deg)", min=None, max=None, round=1),
        SessionAttribute(name="total_manual_picks", title="Total Manual Picks", min=None, max=None, round=0),
        SessionAttribute(name="total_blob_picks", title="Total Blob Picks", min=None, max=None, round=0),
        SessionAttribute(name="blob_pick_score_median", title="Median Blob Pick Score", min=None, max=None, round=3),
        SessionAttribute(name="total_template_picks", title="Total Template Picks", min=None, max=None, round=0),
        SessionAttribute(
            name="template_pick_score_median", title="Median Template Pick Score", min=None, max=None, round=3
        ),
        SessionAttribute(
            name="total_extracted_particles", title="Total Particles Extracted", min=None, max=None, round=0
        ),
        SessionAttribute(
            name="total_extracted_particles_manual",
            title="Total Manual Picker Particles Extracted",
            min=None,
            max=None,
            round=0,
        ),
        SessionAttribute(
            name="total_extracted_particles_blob",
            title="Total Blob Picker Particles Extracted",
            min=None,
            max=None,
            round=0,
        ),
        SessionAttribute(
            name="total_extracted_particles_template",
            title="Total Template Picker Particles Extracted",
            min=None,
            max=None,
            round=0,
        ),
    ]
    """
    List of available session attributes for filtering and display
    """
    picking_thresholds: PickingThresholds = PickingThresholds()
    """
    """
    compute_resources: LiveComputeResources = LiveComputeResources()
    """
    Scheduler compute resource allocation for this session
    """
    notes: str = ""
    """
    User-provided notes about the session
    """
    notes_lock: Optional[str] = None
    """
    Username of user who has locked the notes for editing
    """
    phase_one_wait_for_exposures: bool = False
    """
    Whether to wait until exposures are available before queuing the session's
    preprocessing worker jobs. Set to prevent CryoSPARC from allocating compute
    resources and GPU tokens to workers when the microscope has not yet acquired
    any data.
    """
    phase_one_workers: Dict[str, RtpWorkerState] = {}
    """
    Example: {'J1': RtpWorkerState(status=<JobStatus.BUILDING: 'building'>, errors=[]),
    'J2': RtpWorkerState(status=<JobStatus.RUNNING: 'running'>, errors=[]),
    'J3': RtpWorkerState(status=<JobStatus.KILLED: 'killed'>, errors=[RunError(message='Example', warning=False)])}
    """
    phase_one_workers_soft_kill: List[str] = []
    """
    list of workers that are being signalled a "stop" request
    """
    live_session_job_uid: Optional[str] = None
    """
    UID of the Live session job associated with this session
    """
    file_engine_status: Literal["inactive", "running"] = "inactive"
    """
    Current status of the file engine, which searches for new exposures.
    Inactive while session is paused.
    """
    file_engine_last_run: Optional[float] = None
    """
    Timestamp of the last time the file engine ran
    """
    max_timestamps: List[Any] = []
    """
    Set by the file engine to track the most recent modification timestamps seen in each exposure group
    """
    known_files: List[Any] = []
    """
    :meta private:
    """
    rtp_childs: List[RtpChild] = []
    """
    List of Live worker jobs associated with this session
    """
    template_creation_job: Optional[str] = None
    """
    which job is providing the templates
    """
    template_creation_project: Optional[str] = None
    """
    which project is the above job in
    """
    template_creation_num_particles_in: int = 0
    """
    number of particles provided to template creation
    """
    template_creation_ready: bool = False
    """
    whether the templates are ready to be shown for selection
    """
    template_creation_info: List[TemplateClassInfo] = []
    """
    template info for select 2D (includes gridFS image ids etc)
    """
    exposure_groups: List[ExposureGroup] = []
    """
    Exposure groups configured in this session
    """
    stats: SessionStats = SessionStats()
    """
    """
    data_management: DataManagementStats = DataManagementStats()
    """
    """
    import_signatures: ImportSignature = ImportSignature()
    """
    """
    exposure_summary: Dict[str, Any] = {}
    """
    Summary statistics of exposures processed in this session
    """
    exposure_processing_priority: Literal["normal", "oldest", "latest", "alternate"] = "normal"
    """
    Strategy to use when determining which exposures to process first
    """
    last_compacted_amount: int = 0
    """
    """
    last_compacted_at: Optional[datetime.datetime] = None
    """
    """
    last_compacted_version: Optional[str] = None
    """
    """
    last_restored_amount: int = 0
    """
    """
    last_restored_at: Optional[datetime.datetime] = None
    """
    """
    compacted_exposures_count: int = 0
    """
    """
    restoration_notification_id: Optional[str] = None
    """
    """
    restoration_user_id: Optional[str] = None
    """
    """
    pre_restoration_size: int = 0
    """
    """
    phase2_class2D_restart: bool = False
    """
    (re)start requested - set False by job when it starts
    """
    phase2_class2D_params_spec: LiveClass2DParams = LiveClass2DParams()
    """
    """
    phase2_class2D_params_spec_used: Optional[LiveClass2DParams] = None
    """
    Streaming 2D classification parameters used at last launch, must be same as ``phase2_class2D_params_spec`` to resume
    """
    phase2_class2D_job: Optional[str] = None
    """
    """
    phase2_class2D_ready: bool = False
    """
    """
    phase2_class2D_ready_partial: bool = False
    """
    """
    phase2_class2D_info: List[TemplateClassInfo] = []
    """
    """
    phase2_class2D_num_particles_in: int = 0
    """
    """
    phase2_class2D_particles_out: Optional[Phase2ParticleOutputInfo] = None
    """
    """
    phase2_class2D_num_particles_seen: int = 0
    """
    sum of accepted and rejected
    """
    phase2_class2D_num_particles_accepted: int = 0
    """
    """
    phase2_class2D_num_particles_rejected: int = 0
    """
    """
    phase2_class2D_last_updated: Optional[datetime.datetime] = None
    """
    """
    phase2_select2D_job: Optional[str] = None
    """
    """
    phase2_abinit_restart: bool = False
    """
    """
    phase2_abinit_params_spec: LiveAbinitParams = LiveAbinitParams()
    """
    """
    phase2_abinit_job: Optional[str] = None
    """
    """
    phase2_abinit_ready: bool = False
    """
    """
    phase2_abinit_info: List[LiveVolumeInfo] = []
    """
    """
    phase2_abinit_num_particles_in: int = 0
    """
    """
    phase2_refine_restart: bool = False
    """
    Indicates to the running refinement job that the session parameters have
    changed and a restart is required.
    """
    phase2_refine_params_spec: LiveRefineParams = LiveRefineParams()
    """
    Refinement parameters specified for next launch
    """
    phase2_refine_params_spec_used: Optional[LiveRefineParams] = None
    """
    Refinement parameters used at last launch, must be same as ``phase2_refine_params_spec_used`` to resume
    """
    phase2_refine_job: Optional[str] = None
    """
    """
    phase2_refine_ready: bool = False
    """
    """
    phase2_refine_ready_partial: bool = False
    """
    """
    phase2_refine_num_particles_in: int = 0
    """
    """
    phase2_refine_last_updated: Optional[datetime.datetime] = None
    """
    """
    auto_pause: Literal["disabled", "graceful", "immediate"] = "disabled"
    """
    Auto-pause behavior of the session
    """
    auto_pause_after_idle_minutes: int = 10
    """
    Number of idle minutes before auto-pausing the session
    """
    is_gracefully_pausing: bool = False
    """
    If True, there are no more exposures to process and the session will
    pause once current processing is complete.
    """
    computed_stats_last_run_time: Optional[datetime.datetime] = None
    """
    """
    last_processed_exposure_priority: Literal["normal", "oldest", "latest", "alternate"] = "oldest"
    """
    """
    athena_epu_run_id: Optional[str] = None
    """
    :meta private:
    """
    is_multigrid_epu_run: bool = False
    """
    :meta private:
    """
    ecl: ECLSessionProperties = ECLSessionProperties()
    """
    """
    uid_num: int
    """
    Numeric part of the workspace UID.
    """
    project_uid_num: int
    """
    Numeric part of the project UID.
    """
    session_uid_num: int
    """
    Numeric part of the session UID
    """
    errors: List[SessionBuildError]
    """
    Errors found in the session configuration. Session will not be able
    to start if any errors are present.
    """


class TemplateSelectionThreshold(BaseModel):
    """
    Threshold definition when selecting template classes from 2D classification results
    """

    field: Literal["num_particles_total", "res_A", "class_ess"]
    """
    """
    direction: Literal["above", "below"]
    """
    """
    threshold: Union[int, float]
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, Field

from .file_browser import FileBrowserPrefixes


class Bookmark(BaseModel):
    """
    Bookmarked file path
    """

    id: str
    """
    """
    path: str
    """
    """
    name: str
    """
    """
    description: str
    """
    """
    color: str
    """
    """
    last_accessed: datetime.datetime
    """
    """
    created_at: datetime.datetime
    """
    """


class Email(BaseModel):
    """
    Email address information associated with a user account
    """

    address: str
    """
    """
    verified: bool = False
    """
    """


class RecentPath(BaseModel):
    """
    Recently accessed file path
    """

    path: str
    """
    """
    last_accessed: datetime.datetime
    """
    """


class FileBrowserState(BaseModel):
    """
    State of the user's file browser
    """

    recentPaths: List[RecentPath] = []
    """
    """
    bookmarks: List[Bookmark] = []
    """
    """


class LoginToken(BaseModel):
    """
    Login token for resuming sessions
    """

    hashedToken: str
    """
    """
    when: datetime.datetime
    """
    """


class LoginService(BaseModel):
    """
    Service information for login/resume functionality
    """

    loginTokens: List[LoginToken] = []
    """
    """


class PasswordService(BaseModel):
    """
    Service information for password-based authentication
    """

    bcrypt: str
    """
    The bcrypt-hashed password value
    """


class Profile(BaseModel):
    """
    Profile information for a user account
    """

    name: Union[str, Dict[str, Any]] = ""
    """
    """


class RecentJob(BaseModel):
    """
    Recently accessed job
    """

    project_uid: str
    """
    """
    workspace_uid: str
    """
    """
    job_uid: str
    """
    """


class RecentSession(BaseModel):
    """
    Recently accessed session
    """

    project_uid: str
    """
    """
    session_uid: str
    """
    """


class RecentWorkspace(BaseModel):
    """
    Recently accessed workspace
    """

    project_uid: str
    """
    """
    workspace_uid: str
    """
    """


class Services(BaseModel):
    """
    Authentication services associated with a user account
    """

    password: Optional[PasswordService] = None
    """
    Service information for password-based authentication
    """
    resume: LoginService = LoginService()
    """
    Service information for resuming sessions
    """


class UserState(BaseModel):
    """
    State information for a user account
    """

    licenseAccepted: bool = False
    """
    """
    userCanSetJobPriority: bool = False
    """
    """
    recentProjects: List[str] = []
    """
    """
    recentWorkspaces: List[RecentWorkspace] = []
    """
    """
    recentSessions: List[RecentSession] = []
    """
    """
    recentJobs: List[RecentJob] = []
    """
    """
    browserPath: Optional[str] = None
    """
    """
    defaultJobPriority: Optional[int] = None
    """
    """
    userFileBrowserState: FileBrowserState = FileBrowserState()
    """
    """


class User(BaseModel):
    """
    User account information
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    emails: List[Email] = []
    """
    """
    created_by_user_id: Optional[str] = None
    """
    ID of the admin user who created this account, if applicable
    """
    name: str
    """
    Username
    """
    first_name: str
    """
    First/given name
    """
    last_name: str
    """
    Last/family name
    """
    status: Literal["invited"] = "invited"
    """
    Registration status
    """
    profile: Profile = Profile()
    """
    """
    roles: Dict[str, List[Literal["user", "admin"]]] = {}
    """
    """
    register_token: Optional[str] = None
    """
    """
    reset_token: Optional[str] = None
    """
    """
    last_password_changed_at: Optional[datetime.datetime] = None
    """
    """
    services: Services = Services()
    """
    """
    state: UserState = UserState()
    """
    """
    preferences: Dict[str, Any] = {}
    """
    """
    file_browser_settings: FileBrowserPrefixes = FileBrowserPrefixes()
    """
    """
    lanes: List[str] = []
    """
    List of scheduler lane names that a user is allowed to queue to. Admins
    may queue to all lanes, regardless of this value.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from pydantic import BaseModel


class SystemInfo(BaseModel):
    """
    CryoSPARC general system information.
    """

    version: str
    """
    Running version of CryoSPARC, including any applied patches
    """
    master_hostname: str
    """
    Hostname of the master node
    """
    port_app: int
    """
    Port number for the main CryoSPARC web application
    """
    port_mongo: int
    """
    Port number for the MongoDB database
    """
    port_api: int
    """
    Port number for the CryoSPARC API service
    """
    port_command_vis: int
    """
    Port number for visualization service
    """
    port_rtp_webapp: int
    """
    Port number for real-time API application service.
    """
    is_db_auth_enabled: bool
    """
    Whether authentication is enabled for the database.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, RootModel

from .resource import ResourceSpec

BuilderTag = Literal[
    "new", "interactive", "gpuEnabled", "multiGpu", "utility", "import", "live", "benchmark", "wrapper"
]
"""
Visual indicators for jobs in the builder.
"""

Category = Literal[
    "import",
    "motion_correction",
    "ctf_estimation",
    "exposure_curation",
    "particle_picking",
    "extraction",
    "deep_picker",
    "particle_curation",
    "reconstruction",
    "refinement",
    "ctf_refinement",
    "variability",
    "flexibility",
    "postprocessing",
    "local_refinement",
    "helix",
    "utilities",
    "simulations",
    "live",
    "instance_testing",
    "workflows",
]
"""
Section under which to group a job in the builder.
"""


class InputResult(BaseModel):
    """
    Concrete job input result connection to an output result.
    """

    name: Optional[str] = None
    """
    Input slot name. Passthrough slots have ``name`` set to ``None``.
    """
    dtype: str
    """
    Datatype-specific string from data_registry.py. e.g., stat_blob, ctf,
    alignments2D.
    """
    job_uid: str
    """
    Parent job UID source of this input slot connection.
    """
    output: str
    """
    Name of output in parent job. e.g., "particles"
    """
    result: str
    """
    Name of output result slot in parent job, e.g., "blob". Usually the same
    as "name" but may differ if there are multiple outputs of the same type
    """
    version: Union[int, Literal["F"]] = "F"
    """
    Version number or specifier to use. Usually "F"
    """


class Connection(BaseModel):
    """
    Job input connection details.
    """

    job_uid: str
    """
    Connected parent output job uid.
    """
    output: str
    """
    Name of output on connected parent output job.
    """
    results: List[InputResult] = []
    """
    Specific results from parent job. Some slots may have a different job UID.
    """


class OutputSlot(BaseModel):
    """
    Specification of an output slot in the job configuration. Each output includes one or more.
    """

    name: str
    """
    where to find field in a corresponding .cs file e.g.,
    background_blob
    """
    dtype: str
    """
    Datatype-specific string from data_registry.py. e.g., stat_blob, ctf,
    alignments2D.
    """


class OutputSpec(BaseModel):
    """
    Used for outputs with some generated data based on data forwarded from
    input inheritance
    """

    type: Literal[
        "exposure",
        "particle",
        "template",
        "volume",
        "volume_multi",
        "mask",
        "live",
        "ml_model",
        "symmetry_candidate",
        "flex_mesh",
        "flex_model",
        "hyperparameter",
        "denoise_model",
        "annotation_model",
    ]
    """
    Cryo-EM native data type, e.g., "exposure", "particle" or "volume"
    """
    title: str
    """
    Human-readable title
    """
    description: str = ""
    """
    Detailed description.
    """
    slots: List[Union[OutputSlot, str]] = []
    """
    Expected input/output result slots.

    "str" is a shortcut for ``Slot(dtype="str", "prefix="str")``

    For input specs:
    "str" is a shortcut for ``InputSlot(dtype="str", "prefix="str", required=True)``
    "?str" is a shortcut for ``InputSlot(dtype="str", "prefix="str", required=False)``
    """
    passthrough: Optional[str] = None
    """
    Associated passthrough input name
    """


class OutputRef(BaseModel):
    """
    Minimal name reference to a specific job output
    """

    job_uid: str
    """
    Connected parent output job uid.
    """
    output: str
    """
    Name of output on connected parent output job.
    """


class InputSlot(BaseModel):
    """
    Specification of an input slot in the job configuration. Each input includes one or more.
    """

    name: str
    """
    where to find field in a corresponding .cs file e.g.,
    background_blob
    """
    dtype: str
    """
    Datatype-specific string from data_registry.py. e.g., stat_blob, ctf,
    alignments2D.
    """
    required: bool = False
    """
    whether this field must necessarily exist in acorresponding
    input/output
    """


class Input(BaseModel):
    """
    Job input connection details.
    """

    type: Literal[
        "exposure",
        "particle",
        "template",
        "volume",
        "volume_multi",
        "mask",
        "live",
        "ml_model",
        "symmetry_candidate",
        "flex_mesh",
        "flex_model",
        "hyperparameter",
        "denoise_model",
        "annotation_model",
    ]
    """
    Cryo-EM native data type, e.g., "exposure", "particle" or "volume"
    """
    title: str
    """
    Human-readable title
    """
    description: str = ""
    """
    Detailed description.
    """
    slots: List[InputSlot] = []
    """
    Expected low-level input definitions
    """
    count_min: int = 0
    """
    Minimum number of connections to this input.
    """
    count_max: Union[int, Literal["inf"]] = "inf"
    """
    Maximum number of connections supported for this input. Should be any
    integer >= 0 and <= 500. Inputs with a ``count_max`` set to ``"inf"`` also
    support a maximum of 500 connections.
    """
    repeat_allowed: bool = False
    """
    Whether repeated connections to the same output allowed for this input.
    """
    connections: List[Connection] = []
    """
    Connected output details
    """


class InputSpec(BaseModel):
    """
    Input specification. Used to define the expected connections to a job input.
    """

    type: Literal[
        "exposure",
        "particle",
        "template",
        "volume",
        "volume_multi",
        "mask",
        "live",
        "ml_model",
        "symmetry_candidate",
        "flex_mesh",
        "flex_model",
        "hyperparameter",
        "denoise_model",
        "annotation_model",
    ]
    """
    Cryo-EM native data type, e.g., "exposure", "particle" or "volume"
    """
    title: str
    """
    Human-readable title
    """
    description: str = ""
    """
    Detailed description.
    """
    slots: List[Union[InputSlot, str]] = []
    """
    Expected input/output result slots.

    "str" is a shortcut for ``Slot(dtype="str", "prefix="str")``

    For input specs:
    "str" is a shortcut for ``InputSlot(dtype="str", "prefix="str", required=True)``
    "?str" is a shortcut for ``InputSlot(dtype="str", "prefix="str", required=False)``
    """
    count_min: int = 0
    """
    Minimum number of connections to this input.
    """
    count_max: Union[int, Literal["inf"]] = "inf"
    """
    Maximum number of connections supported for this input. Should be any
    integer >= 0 and <= 500. Inputs with a ``count_max`` set to ``"inf"`` also
    support a maximum of 500 connections.
    """
    repeat_allowed: bool = False
    """
    Whether repeated connections to the same output allowed for this input.
    """


class InputSpecs(RootModel):
    """
    Dictionary of input specifications, where each key is the input name.
    """

    root: Dict[str, InputSpec] = {}
    """
    Dictionary of input specifications, where each key is the input name.
    """


class Inputs(RootModel):
    """
    Dictionary of job input connection details, where each key is the input name.
    """

    root: Dict[str, Input] = {}
    """
    Dictionary of job input connection details, where each key is the input name.
    """


class Params(BaseModel):
    """
    Job parameter specifications. See API function projects.get_job_register
    for allowed parameters based on job spec type.
    """

    model_config = ConfigDict(extra="allow")
    if TYPE_CHECKING:

        def __init__(self, **kwargs: Any) -> None: ...
        def __getattr__(self, key: str) -> Any: ...


class OutputResult(BaseModel):
    """
    Concrete job output.
    """

    name: str
    """
    where to find field in a corresponding .cs file e.g.,
    background_blob
    """
    dtype: str
    """
    Datatype-specific string from data_registry.py. e.g., stat_blob, ctf,
    alignments2D.
    """
    versions: List[int] = []
    """
    List of available intermediate result version numbers.
    """
    metafiles: List[str] = []
    """
    List of available intermediate result files (same size as ``versions``).
    """
    num_items: List[int] = []
    """
    Number of rows in each metafile
    """
    passthrough: bool = False
    """
    If True, this result is passed through as-is from an associated input.
    """


class Output(BaseModel):
    """
    Job output details. Includes saved dataset paths and summary statistics.
    """

    type: Literal[
        "exposure",
        "particle",
        "template",
        "volume",
        "volume_multi",
        "mask",
        "live",
        "ml_model",
        "symmetry_candidate",
        "flex_mesh",
        "flex_model",
        "hyperparameter",
        "denoise_model",
        "annotation_model",
    ]
    """
    Cryo-EM native data type, e.g., "exposure", "particle" or "volume"
    """
    title: str
    """
    Human-readable title
    """
    description: str = ""
    """
    Detailed description.
    """
    slots: List[OutputSlot] = []
    """
    Low-level output result definitions.
    """
    passthrough: Optional[str] = None
    """
    Associated passthrough input name
    """
    results: List[OutputResult] = []
    """
    All individial outputs based on available output slots
    """
    num_items: int = 0
    """
    Number of items in final version file
    """
    image: Optional[str] = None
    """
    Asset ID of thumbnail for this output
    """
    summary: Dict[str, Any] = {}
    """
    Result dataset summary data
    """
    latest_summary_stats: Dict[str, Any] = {}
    """
    Additional context-specific summary statistics
    """


class Outputs(RootModel):
    """
    Dictionary of job output details, where each key is the output name.
    """

    root: Dict[str, Output] = {}
    """
    Dictionary of job output details, where each key is the output name.
    """


class JobSpec(BaseModel):
    """
    Job's unique specification details. Defines the parameters, inputs, outputs
    and required resources. Contents vary by job type, configured parameters and
    connected inputs.
    """

    type: str
    """
    Job type key, e.g., "import_particles" or "class_2D_new"
    """
    params: Params = Params()
    """
    Parameters for job, attributes vary per job type.

    NOTE: After changing a job parameter, the spec may need to be refreshed.
    Instead of directly modifying this field with job.params.foo = ... ,
    use job.set_param("foo", ...) instead.
    """
    inputs: Inputs = Inputs()
    """
    Connected inputs
    """
    outputs: Outputs = Outputs()
    """
    Available outputs. Empty when job has not run.
    """
    ui_tile_width: int
    """
    Number of horizontal tiles this job should take in the UI.
    """
    ui_tile_height: int
    """
    Number of vertical tiles this job should take in the UI.
    """
    resource_spec: ResourceSpec
    """
    Compute resource requirements for this job.
    """


class JobBuildError(BaseModel):
    type: Literal[
        "no_such_attribute",
        "json_invalid",
        "json_type",
        "needs_python_object",
        "recursion_loop",
        "missing",
        "frozen_field",
        "frozen_instance",
        "extra_forbidden",
        "invalid_key",
        "get_attribute_error",
        "model_type",
        "model_attributes_type",
        "dataclass_type",
        "dataclass_exact_type",
        "none_required",
        "greater_than",
        "greater_than_equal",
        "less_than",
        "less_than_equal",
        "multiple_of",
        "finite_number",
        "too_short",
        "too_long",
        "iterable_type",
        "iteration_error",
        "string_type",
        "string_sub_type",
        "string_unicode",
        "string_too_short",
        "string_too_long",
        "string_pattern_mismatch",
        "enum",
        "dict_type",
        "mapping_type",
        "list_type",
        "tuple_type",
        "set_type",
        "set_item_not_hashable",
        "bool_type",
        "bool_parsing",
        "int_type",
        "int_parsing",
        "int_parsing_size",
        "int_from_float",
        "float_type",
        "float_parsing",
        "bytes_type",
        "bytes_too_short",
        "bytes_too_long",
        "bytes_invalid_encoding",
        "value_error",
        "assertion_error",
        "literal_error",
        "date_type",
        "date_parsing",
        "date_from_datetime_parsing",
        "date_from_datetime_inexact",
        "date_past",
        "date_future",
        "time_type",
        "time_parsing",
        "datetime_type",
        "datetime_parsing",
        "datetime_object_invalid",
        "datetime_from_date_parsing",
        "datetime_past",
        "datetime_future",
        "timezone_naive",
        "timezone_aware",
        "timezone_offset",
        "time_delta_type",
        "time_delta_parsing",
        "frozen_set_type",
        "is_instance_of",
        "is_subclass_of",
        "callable_type",
        "union_tag_invalid",
        "union_tag_not_found",
        "arguments_type",
        "missing_argument",
        "unexpected_keyword_argument",
        "missing_keyword_only_argument",
        "unexpected_positional_argument",
        "missing_positional_only_argument",
        "multiple_argument_values",
        "url_type",
        "url_parsing",
        "url_syntax_violation",
        "url_too_long",
        "url_scheme",
        "uuid_type",
        "uuid_parsing",
        "uuid_version",
        "decimal_type",
        "decimal_parsing",
        "decimal_max_digits",
        "decimal_max_places",
        "decimal_whole_digits",
        "complex_type",
        "complex_str_parsing",
    ]
    """
    values based on https://docs.pydantic.dev/latest/errors/validation_errors
    """
    loc: List[Union[str, int]]
    """
    path to the invalid property
    """
    input: Any
    """
    value of the invalid property - must be serializable
    """
    ctx: Dict[str, Any] = {}
    """
    error context for pydantic
    """
    input_type: str
    """
    """


Stability = Literal["develop", "beta", "stable", "legacy", "obsolete"]
"""
Lifecycle/development stage for a job. Jobs will change stabilities as they
are continually developed or replaced with other jobs.
"""


class OutputSpecs(RootModel):
    """
    Dictionary of output specifications, where each key is the output name.
    """

    root: Dict[str, OutputSpec] = {}
    """
    Dictionary of output specifications, where each key is the output name.
    """


class JobRegisterError(BaseModel):
    """
    Error that occurs when loading a developer job register.
    """

    type: str
    """
    """
    message: str
    """
    """
    traceback: str
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import List, Optional

from pydantic import BaseModel

from .gpu import GpuInfo


class InstanceInformation(BaseModel):
    """
    System information for a machine running a CryoSPARC master or worker installation.
    """

    platform_node: str = ""
    """
    Node hostname
    """
    platform_release: str = ""
    """
    Linux kernel release version
    """
    platform_version: str = ""
    """
    Platform version string
    """
    platform_architecture: str = ""
    """
    Platform architecture (e.g., x86_64)
    """
    cpu_model: str = ""
    """
    CPU model name
    """
    physical_cores: int = 0
    """
    Number of physical CPU cores
    """
    max_cpu_freq: float = 0.0
    """
    Maximum CPU frequency in MHz
    """
    total_memory: str = "0B"
    """
    Total system memory, formatted as a string with units (e.g., "1.23GB")
    """
    available_memory: str = "0B"
    """
    Available system memory, formatted as a string with units (e.g., "1.23GB")
    """
    used_memory: str = "0B"
    """
    Used system memory, formatted as a string with units (e.g., "1.23GB")
    """
    ofd_soft_limit: int = 0
    """
    Maximum allowed open file descriptors (soft limit)
    """
    ofd_hard_limit: int = 0
    """
    Maximum allowed open file descriptors (hard limit)
    """
    driver_version: Optional[str] = None
    """
    Maximum CUDA version supported by Nvidia driver (e.g., "12.9")
    """
    toolkit_version: Optional[str] = None
    """
    Installed CUDA toolkit version (e.g., "12.8")
    """
    CUDA_version: Optional[str] = None
    """
    same as toolkit_version, for backwards compatibility :meta private:
    """
    nvrtc_version: Optional[str] = None
    """
    Installed NVRTC version (e.g., "12.8")
    """
    gpu_info: Optional[List[GpuInfo]] = None
    """
    List of GPU information objects
    """
    version: str = ""
    """
    Installed CryoSPARC version
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field


class JobGroup(BaseModel):
    """
    User-defined group of jobs within a workspace.
    """

    id: int
    """
    """
    jobs: List[str]
    """
    """
    title: Optional[str] = None
    """
    """
    description: Optional[str] = None
    """
    """
    color: Optional[str] = None
    """
    """


class WorkspaceStats(BaseModel):
    """
    Statistics about the contents of a workspace.
    """

    updated_at: Optional[datetime.datetime] = None
    """
    """
    job_count: int = 0
    """
    """
    job_sections: Dict[str, int] = {}
    """
    Number of jobs per section.
    """
    job_status: Dict[str, int] = {}
    """
    Number of jobs per status.
    """
    job_types: Dict[str, int] = {}
    """
    Number of jobs per type.
    """


class WorkspaceLastAccessed(BaseModel):
    """
    Record for when a user last accessed a workspace.
    """

    name: str
    """
    Username that last accessed the workspace.
    """
    accessed_at: datetime.datetime
    """
    Timestamp of the last access.
    """


class Workspace(BaseModel):
    """
    A workspace within a project, which contains jobs and related data. Jobs
    are stored in a flat structure within the project and linked to one or more
    workspaces.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    dumped_at: Optional[datetime.datetime] = None
    """
    When the model was last dumped to disk
    """
    last_dumped_version: Optional[str] = None
    """
    The version of CryoSPARC last dumped at
    """
    autodump: bool = True
    """
    Whether the model was updated recently and must be dumped

    :meta private:
    """
    uid: str
    """
    Workspace unique ID, e.g., 'W1'.
    """
    project_uid: str
    """
    Project unique ID for , e.g., 'P1'.
    """
    workspace_type: Literal["base", "live"] = "base"
    """
    Workspace type, either 'base' or 'live'.
    """
    created_by_user_id: Optional[str] = None
    """
    User ID that created the workspace.
    """
    title: Optional[str] = None
    """
    """
    description: Optional[str] = None
    """
    """
    created_by_job_uid: Optional[str] = None
    """
    """
    tags: List[str] = []
    """
    """
    starred_by: List[str] = []
    """
    List of user IDs who have starred the workspace.
    """
    deleted: bool = False
    """
    """
    deleting: bool = False
    """
    """
    last_accessed: Optional[WorkspaceLastAccessed] = None
    """
    """
    workspace_stats: WorkspaceStats = WorkspaceStats()
    """
    """
    imported_at: Optional[datetime.datetime] = None
    """
    """
    groups: List[JobGroup] = []
    """
    """
    uid_num: int
    """
    Numeric part of the workspace UID.
    """
    project_uid_num: int
    """
    Numeric part of the project UID.
    """

    model_config = ConfigDict(extra="allow")
    if TYPE_CHECKING:

        def __init__(self, **kwargs: Any) -> None: ...
        def __getattr__(self, key: str) -> Any: ...



cdef extern from "lz4/lib/lz4.h":
    int LZ4_compress_default(const char *src, char *dst, int srcSize, int dstCapacity) nogil
    int LZ4_decompress_safe (const char *src, char *dst, int compressedSize, int dstCapacity) nogil
    int LZ4_compressBound(int inputSize) nogil
    int LZ4_compress_fast (const char *src, char *dst, int srcSize, int dstCapacity, int acceleration) nogil



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, Field

from .instance import InstanceInformation


class PerformanceBenchmark(BaseModel):
    """
    Performance benchmark results a benchmarking job.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    project_uid: str
    """
    """
    job_uid: str
    """
    """
    type: Literal["fs", "cpu", "gpu", "extensive_validation"]
    """
    """
    cryosparc_version: str
    """
    """
    instance_information: InstanceInformation
    """
    """
    job_params: Dict[str, Any] = {}
    """
    """
    timings: Dict[str, Any] = {}
    """
    """
    gpu_name: Optional[str] = None
    """
    """
    gpu_bus_id: Optional[str] = None
    """
    """
    sequential_read_benchmark_dir: Optional[str] = None
    """
    """
    sequential_disk_models: List[Dict[str, Any]] = []
    """
    """
    particle_read_benchmark_dir: Optional[str] = None
    """
    """
    random_disk_models: List[Dict[str, Any]] = []
    """
    """


class ReferencePerformanceBenchmark(BaseModel):
    """
    Reference performance benchmark for comparison purposes.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    type: Literal["fs", "cpu", "gpu", "extensive_validation"]
    """
    """
    title: str
    """
    """
    cryosparc_version: str
    """
    """
    instance_information: InstanceInformation = InstanceInformation()
    """
    """
    job_params: Dict[str, Any] = {}
    """
    """
    timings: Dict[str, Any] = {}
    """
    """
    gpu_name: Optional[str] = None
    """
    """
    gpu_bus_id: Optional[str] = None
    """
    """
    reference_platform_node: Optional[str] = None
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from pydantic import BaseModel


class ImportSignature(BaseModel):
    """
    Binary signatures of imported paths used in import jobs and sessions,
    to track re-importing the same data
    """

    count: int = 0
    """
    Number of imported paths
    """
    signatures: bytes = b""
    """
    Binary signatures of imported paths
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Dict, List, Literal, Optional, Union

from pydantic import BaseModel

from .gpu import Gpu
from .resource import FixedResourceSlots, ResourceSlots


class Node(BaseModel):
    """
    Node-type scheduler target that may include GPUs
    """

    type: Literal["node"]
    """
    Node scheduler targets have type "node".
    """
    ssh_str: str
    """
    Shell command used to access this node, e.g., ``ssh cryosparcuser@worker``.
    """
    resource_slots: ResourceSlots = ResourceSlots()
    """
    Available compute resources.
    """
    resource_fixed: FixedResourceSlots = FixedResourceSlots()
    """
    Available fixed resources.
    """
    monitor_port: Optional[int] = None
    """
    Not used.
    """
    gpus: Optional[List[Gpu]] = None
    """
    Details for all GPUs available on this node, including those excluded from ``resource_slots``.
    """


class Cluster(BaseModel):
    """
    Cluster-type scheduler targets details
    """

    send_cmd_tpl: str = "{{ command }}"
    """
    Template command to access the cluster and running commands.
    """
    qsub_cmd_tpl: str = "qsub {{ script_path_abs }}"
    """
    Template command to submit jobs to the cluster.
    """
    qstat_cmd_tpl: str = "qstat -as {{ cluster_job_id }}"
    """
    Template command to check the cluster job by its ID.
    """
    qdel_cmd_tpl: str = "qdel {{ cluster_job_id }}"
    """
    Template command to delete cluster jobs.
    """
    qinfo_cmd_tpl: str = "qstat -q"
    """
    Template command to check cluster queue info.
    """
    type: Literal["cluster"]
    """
    Cluster scheduler targets have type "cluster".
    """
    script_tpl: str = ""
    """
    Full cluster submission script Jinja template.
    """
    custom_vars: Dict[str, str] = {}
    """
    Custom variable values
    """
    tpl_vars: List[str]
    """
    List of template variable names in a cluster target
    """
    custom_var_names: List[str]
    """
    Computed list if custom variable names defined in the template
    """


class SchedulerTarget(BaseModel):
    """
    Details and configuration for a node or cluster target.
    """

    cache_path: Optional[str] = None
    """
    Path the SSD cache scratch directory, if applicable.
    """
    cache_reserve_mb: Optional[int] = None
    """
    Ensure at least this much space is free on the SSD scratch drive before
    caching.
    """
    cache_quota_mb: Optional[int] = None
    """
    Do not cache more than this amount on the SSD scrath drive..
    """
    lane: str
    """
    Lane name this target belongs to.
    """
    name: str
    """
    Identifier for this target.
    """
    title: str
    """
    Human-readable title for this target.
    """
    desc: Optional[str] = None
    """
    Human-readable description for this target.
    """
    hostname: str
    """
    Network machine hostname (same as name for for clusters).
    """
    worker_bin_path: str
    """
    Path to cryosparc_worker/bin/cryosparcw executable.
    """
    config: Union[Node, Cluster]
    """
    Target configuration details.
    """


class SchedulerTargetCluster(BaseModel):
    cache_path: Optional[str] = None
    """
    Path the SSD cache scratch directory, if applicable.
    """
    cache_reserve_mb: Optional[int] = None
    """
    Ensure at least this much space is free on the SSD scratch drive before
    caching.
    """
    cache_quota_mb: Optional[int] = None
    """
    Do not cache more than this amount on the SSD scrath drive..
    """
    lane: str
    """
    Lane name this target belongs to.
    """
    name: str
    """
    Identifier for this target.
    """
    title: str
    """
    Human-readable title for this target.
    """
    desc: Optional[str] = None
    """
    Human-readable description for this target.
    """
    hostname: str
    """
    Network machine hostname (same as name for for clusters).
    """
    worker_bin_path: str
    """
    Path to cryosparc_worker/bin/cryosparcw executable.
    """
    config: Cluster
    """
    Target configuration details.
    """


class SchedulerTargetNode(BaseModel):
    cache_path: Optional[str] = None
    """
    Path the SSD cache scratch directory, if applicable.
    """
    cache_reserve_mb: Optional[int] = None
    """
    Ensure at least this much space is free on the SSD scratch drive before
    caching.
    """
    cache_quota_mb: Optional[int] = None
    """
    Do not cache more than this amount on the SSD scrath drive..
    """
    lane: str
    """
    Lane name this target belongs to.
    """
    name: str
    """
    Identifier for this target.
    """
    title: str
    """
    Human-readable title for this target.
    """
    desc: Optional[str] = None
    """
    Human-readable description for this target.
    """
    hostname: str
    """
    Network machine hostname (same as name for for clusters).
    """
    worker_bin_path: str
    """
    Path to cryosparc_worker/bin/cryosparcw executable.
    """
    config: Node
    """
    Target configuration details.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from pydantic import BaseModel, Field

from .instance import InstanceInformation
from .job_spec import JobBuildError, JobSpec
from .resource import FixedResourceSlots, ResourceSlots
from .scheduler_target import SchedulerTarget


class AllocatedResources(BaseModel):
    """
    Scheduler target resources allocated for a job the scheduler at launch time.
    """

    lane: Optional[str] = None
    """
    Name of lane the job is queued to.
    """
    lane_type: Optional[str] = None
    """
    Type of allocated lane ("node" or "cluster")
    """
    hostname: str
    """
    For node targets, hostname of selected worker node.
    """
    target: Optional[SchedulerTarget] = None
    """
    Allocated scheduler target full details.
    """
    slots: ResourceSlots = ResourceSlots()
    """
    Listings of compute resources allocated for this job.
    """
    fixed: FixedResourceSlots = FixedResourceSlots()
    """
    Fixed compute resources allocated for this job.
    """
    licenses_acquired: int = 0
    """
    Number of GPU tokens allocated for this job.
    """


JobStatus = Literal["building", "queued", "launched", "started", "running", "waiting", "completed", "killed", "failed"]


class JobLastAccessed(BaseModel):
    """
    Record for when a user last accessed a job.
    """

    name: str
    """
    Username that last accessed the job.
    """
    accessed_at: datetime.datetime = datetime.datetime(1, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    Timestamp of the last access.
    """


class RunError(BaseModel):
    """
    Details about a runtime error encountered during job execution.
    """

    message: str
    """
    """
    warning: bool = False
    """
    """


class UiTileImage(BaseModel):
    """
    Details about a job tile image shown in the UI.
    """

    name: str
    """
    """
    fileid: str
    """
    """
    num_rows: Optional[int] = None
    """
    """
    num_cols: Optional[int] = None
    """
    """


class JobWorkflowInfo(BaseModel):
    """
    Information about the workflow a job is part of.
    """

    id: str
    """
    """
    jobId: str
    """
    """
    run: int = 0
    """
    """


class Job(BaseModel):
    """
    Specification for a CryoSPARC Job object. Access parameters, inputs and
    outputs from `spec` field.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    dumped_at: Optional[datetime.datetime] = None
    """
    When the model was last dumped to disk
    """
    last_dumped_version: Optional[str] = None
    """
    The version of CryoSPARC last dumped at
    """
    autodump: bool = True
    """
    Whether the model was updated recently and must be dumped

    :meta private:
    """
    uid: str
    """
    Job unique ID, e.g., "J42".
    """
    project_uid: str
    """
    Project unique ID, e.g., "P3".
    """
    workspace_uids: List[str] = ["W1"]
    """
    List of workspace UIDs this job belongs to. Must have at least one item
    """
    spec: JobSpec
    """
    Job type-specific settings, including params, inputs, outputs, resources, etc.
    """
    job_dir: str
    """
    Job directory relative to project directory, usually the same as uid
    """
    job_dir_size: int = 0
    """
    Job directory size
    """
    job_dir_size_last_updated: datetime.datetime
    """
    Job directory last updated timestamp
    """
    run_as_user: Optional[str] = None
    """
    Developer job register to use (internal use only)
    """
    title: str = ""
    """
    Human-readable job title.
    """
    description: str = ""
    """
    Human-readable job markdown description.
    """
    status: JobStatus = "building"
    """
    Job scheduling status, e.g., "building", "queued", "running".
    """
    created_by_user_id: Optional[str] = None
    """
    Object ID of user account that created this job.
    """
    created_by_job_uid: Optional[str] = None
    """
    Job ID of job that created this job.
    """
    cloned_from: Optional[str] = None
    """
    Job this job was cloned from
    """
    queued_at: Optional[datetime.datetime] = None
    """
    When added to the queue
    """
    started_at: Optional[datetime.datetime] = None
    """
    When internal scheduler launches/submits
    """
    launched_at: Optional[datetime.datetime] = None
    """
    When job initial monitor process starts
    """
    running_at: Optional[datetime.datetime] = None
    """
    When job monitor process forks into main process
    """
    waiting_at: Optional[datetime.datetime] = None
    """
    When job last went into waiting for interaction status
    """
    completed_at: Optional[datetime.datetime] = None
    """
    When job finished
    """
    killed_at: Optional[datetime.datetime] = None
    """
    When job was killed
    """
    failed_at: Optional[datetime.datetime] = None
    """
    When job failed
    """
    heartbeat_at: Optional[datetime.datetime] = None
    """
    Last heartbeat received
    """
    tokens_acquired_at: Optional[datetime.datetime] = None
    """
    When job got confirmation from the licensing server to run
    """
    tokens_requested_at: Optional[datetime.datetime] = None
    """
    When job started waiting for tokens to become available
    """
    last_scheduled_at: Optional[datetime.datetime] = None
    """
    When scheduler last tried to launch this job
    """
    last_accessed: Optional[JobLastAccessed] = None
    """
    When a specific user last opened this job in the user interface
    """
    has_error: bool = False
    """
    If True, there's an error in the streamlog
    """
    has_warning: bool = False
    """
    If True, there's a warning in the streamlog
    """
    version_created: Optional[str] = None
    """
    Version of CryoSPARC used to create this job
    """
    version: Optional[str] = None
    """
    Version of CryoSPARC that last ran this job
    """
    priority: int = 0
    """
    Queue priority set by user or default for system. higher number = higher priority.
    """
    deleted: bool = False
    """
    True if the job has been marked as deleted.
    """
    deleting: bool = False
    """
    True if the job is being deleted
    """
    parents: List[str] = []
    """
    Set of parent jobs UIDs based on input connections.
    """
    children: List[str] = []
    """
    Set of child job UIDs based on output connections.
    """
    resources_allocated: Optional[AllocatedResources] = None
    """
    Set by scheduler before launching
    """
    queued_by_user_id: Optional[str] = None
    """
    ID of user account that queued this job.
    """
    queued_to_lane: Optional[str] = None
    """
    set at queue time based on params
    """
    queued_to_hostname: Optional[str] = None
    """
    NOTE: database field is sometimes ``False`` in older CryoSPARC jobs, cast validator prevents this
    """
    queued_to_gpu: Optional[List[int]] = None
    """
    NOTE: database field is sometimes ``False`` in older CryoSPARC jobs, cast validator prevents this
    """
    queue_status: Optional[
        Literal[
            "waiting_inputs",
            "project_paused",
            "actively_queued",
            "launched",
            "waiting_maintenance",
            "waiting_licenses",
            "waiting_resources",
        ]
    ] = None
    """
    waiting_inputs, project_paused, actively_queued, launched
    """
    queue_message: Optional[str] = None
    """
    e.g., "Waiting for resource in lane xxx: GPU"
    """
    queued_job_hash: Optional[int] = None
    """
    hash of puid, juid, # tokens requested, job type and user for scheduler
    """
    num_tokens: int = 0
    """
    number of (GPU) tokens required to run the job
    """
    job_sig: Optional[str] = None
    """
    Job signature created by licensing module
    """
    errors_run: List[RunError] = []
    """
    Runtime errors, limited to last 10 errors.
    """
    interactive_port: Optional[int] = None
    """
    Port assigned to a job's interactive server by operation system, only
    available after an interactive job enters waiting status.
    """
    PID_monitor: Optional[int] = None
    """
    Process ID number of job monitor process
    """
    PID_main: Optional[int] = None
    """
    Process ID number of main job process
    """
    PID_workers: List[int] = []
    """
    Process ID numbers of worker processes spawned by main job process
    """
    cluster_job_id: Optional[str] = None
    """
    """
    cluster_job_status: Optional[str] = None
    """
    """
    cluster_job_status_code: Optional[str] = None
    """
    """
    cluster_job_monitor_event_id: Optional[str] = None
    """
    """
    cluster_job_monitor_retries: int = 0
    """
    """
    cluster_job_monitor_last_run_at: Optional[datetime.datetime] = None
    """
    """
    cluster_job_submission_script: Optional[str] = None
    """
    """
    cluster_job_custom_vars: Dict[str, str] = {}
    """
    """
    ui_tile_images: List[UiTileImage] = []
    """
    Job tile images in app
    """
    is_experiment: bool = False
    """
    Is the job part of an experiment?
    """
    enable_bench: bool = False
    """
    Enable benchmarking for this job. Used by extensive workflow
    """
    bench: Dict[str, Union[float, int, Dict[str, Any]]] = {}
    """
    Benchmarking data
    """
    bench_timings: Dict[str, List[Tuple[datetime.datetime, Optional[datetime.datetime]]]] = {}
    """
    Benchmark timing data in the form {key: list_of_timings}
    """
    completed_count: int = 0
    """
    How many times this job was marked as completed
    """
    instance_information: InstanceInformation = InstanceInformation()
    """
    Details about the active instance
    """
    last_intermediate_data_cleared_at: Optional[datetime.datetime] = None
    """
    """
    last_intermediate_data_cleared_amount: int = 0
    """
    """
    intermediate_results_size_bytes: int = 0
    """
    """
    intermediate_results_size_last_updated: Optional[datetime.datetime] = None
    """
    """
    is_final_result: bool = False
    """
    Job and its anscestors is the final result and should not be deleted
    during data cleanup steps.
    """
    is_ancestor_of_final_result: bool = False
    """
    Similar to is_final_result
    """
    no_check_inputs_ready: bool = False
    """
    If True, a queued job will be launched even if parent connected jobs have
    not finished running yet.
    """
    progress: List[Dict[str, Any]] = []
    """
    Progress log
    """
    last_exported_at: Optional[datetime.datetime] = None
    """
    Time of last export of outputs
    """
    last_exported_location: Optional[str] = None
    """
    Time of last exported outputs location
    """
    last_exported_version: Optional[str] = None
    """
    Version of CryoSPARC this job was last exported from
    """
    tags: List[str] = []
    """
    Tags associated with this job
    """
    workflow: Optional[JobWorkflowInfo] = None
    """
    Information about workflow that job ran from, if any
    """
    imported_at: Optional[datetime.datetime] = None
    """
    """
    deleted_at: Optional[datetime.datetime] = None
    """
    """
    import_status: Optional[Literal["importing", "complete", "failed"]] = None
    """
    """
    attach_status: Optional[Literal["attaching", "complete", "failed"]] = None
    """
    """
    starred_by: List[str] = []
    """
    List of user IDs who have starred this job.
    """
    requeue_windows_ends_at: Optional[datetime.datetime] = None
    """
    If set, job can be cleared and requeued until this time without losing its place in the queue
    """
    uid_num: int
    """
    Numeric part of the job UID.
    """
    project_uid_num: int
    """
    Numeric part of the project UID.
    """
    status_num: int
    """
    Numeric representation of job status.
    """
    build_errors: List[JobBuildError]
    """
    List of build errors for this job. Jobs cannot be queued or launched if there are build errors.
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import Literal

from pydantic import BaseModel


class Token(BaseModel):
    """
    Login token information for API authentication.
    """

    access_token: str
    """
    """
    token_type: Literal["bearer"]
    """
    """



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
import datetime
from typing import Literal, Optional

from pydantic import BaseModel, Field


class TagCounts(BaseModel):
    """
    Counts of tagged items by type.
    """

    total: int = 0
    """
    """
    project: int = 0
    """
    """
    workspace: int = 0
    """
    """
    session: int = 0
    """
    """
    job: int = 0
    """
    """


class Tag(BaseModel):
    """
    A user-defined tag that can be applied to various entities, including
    projects, workspaces, sessions and jobs.
    """

    id: str = Field("000000000000000000000000", alias="_id")
    """
    """
    updated_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was last modified.
    """
    created_at: datetime.datetime = datetime.datetime(1970, 1, 1, 0, 0, tzinfo=datetime.timezone.utc)
    """
    When this object was first created. Imported objects such as projects
    and jobs will retain the created time from their original CryoSPARC instance.
    """
    uid: str
    """
    """
    title: str
    """
    """
    type: Literal["general", "project", "workspace", "session", "job"]
    """
    """
    created_by_user_id: str
    """
    """
    colour: Optional[
        Literal[
            "black",
            "gray",
            "red",
            "orange",
            "yellow",
            "green",
            "teal",
            "cyan",
            "sky",
            "blue",
            "indigo",
            "purple",
            "pink",
        ]
    ] = "gray"
    """
    """
    description: Optional[str] = None
    """
    """
    created_by_workflow: Optional[str] = None
    """
    """
    counts: TagCounts = TagCounts()
    """
    """
    uid_num: int
    """
    Numeric part of the tag UID.
    """



"""
Core base classes and utilities for other cryosparc-tools modules.
"""

# NOTE: This file should only include utilities required only by cryosparc-tools
# CryoSPARC should not depend on anything in this file.
import warnings
from abc import ABC, abstractmethod
from typing import Any, Dict, Generic, Optional, TypeVar, Union

from pydantic import BaseModel

from ..models.job_spec import InputSlot, OutputSlot
from ..spec import SlotSpec

# API model
M = TypeVar("M", bound=BaseModel)


class Controller(ABC, Generic[M]):
    """
    Abstract base class for Project, Workspace, Job classes and any other types
    that have underlying Mongo database documents.

    Generic type argument D is a typed dictionary definition for a Mongo
    document.

    :meta private:
    """

    _model: Optional[M] = None

    @property
    def model(self) -> M:
        """
        Representation of entitity data. Contents may change in CryoSPARC
        over time, use use :py:meth:`refresh` to update.
        """
        if not self._model:
            self.refresh()
        assert self._model, "Could not refresh database document"
        return self._model

    @model.setter
    def model(self, model: M):
        self._model = model

    @model.deleter
    def model(self):
        self._model = None

    @property
    def doc(self) -> Dict[str, Any]:
        warnings.warn(".doc attribute is deprecated. Use .model attribute instead.", DeprecationWarning, stacklevel=2)
        return self.model.model_dump(by_alias=True)

    @abstractmethod
    def refresh(self):
        # Must be implemented in subclasses
        return self


def as_input_slot(spec: Union[SlotSpec, InputSlot]) -> InputSlot:
    if isinstance(spec, str):
        spec, required = (spec[1:], False) if spec[0] == "?" else (spec, True)
        return InputSlot(name=spec, dtype=spec, required=required)
    elif isinstance(spec, dict) and "dtype" in spec:
        dtype = spec["dtype"]
        name = spec.get("name") or spec.get("prefix") or dtype
        required = spec.get("required", True)
        return InputSlot(name=name, dtype=dtype, required=required)
    return spec


def as_output_slot(spec: Union[SlotSpec, OutputSlot]) -> OutputSlot:
    if isinstance(spec, str):
        return OutputSlot(name=spec, dtype=spec)
    elif isinstance(spec, dict) and "dtype" in spec:
        dtype = spec["dtype"]
        name = spec.get("name") or spec.get("prefix") or dtype
        return OutputSlot(name=name, dtype=dtype)
    return spec



# THIS FILE IS AUTO-GENERATED, DO NOT EDIT DIRECTLY
# SEE dev/api_generate_models.py
from typing import List

from pydantic import BaseModel


class GetFinalResultsResponse(BaseModel):
    """
    Information about jobs marked as final in a project.
    """

    final_results: List[str]
    """
    Job uids of jobs marked as final result.
    """
    ancestors_of_final_results: List[str]
    """
    Job uids of ancestors of final result jobs.
    """
    non_ancestors_of_final_results: List[str]
    """
    All other job uids in the project that are neither marked as final nor ancestors of final result jobs.
    """


class Hello(BaseModel):
    """
    Response model for root endpoint
    """

    name: str = "CryoSPARC"
    """
    """
    version: str
    """
    Running CryoSPARC version. Includes patch, if installed
    """
    service: str
    """
    API service name
    """


class WorkspaceAncestorUidsResponse(BaseModel):
    """
    Listings of ancestor and non-ancestor jobs in a given workspace.
    """

    ancestors: List[str]
    """
    Jobs in the workspace that are ancestors of a given set of jobs.
    """
    non_ancestors: List[str]
    """
    Jobs in the workspace that are not ancestors of a given set of jobs.
    """


class WorkspaceDescendantUidsResponse(BaseModel):
    """
    Listings of descendant and non-descendant jobs in a given workspace.
    """

    descendants: List[str]
    """
    Jobs in the workspace that are descendants of a given set of jobs.
    """
    non_descendants: List[str]
    """
    Jobs in the workspace that are not descendants of a given set of jobs.
    """
